package jig.engine.physics.vpe;

import java.awt.Color;
import java.util.List;

import jig.engine.RenderingContext;
import jig.engine.ShapeEngine;
import jig.engine.util.Matrix22;
import jig.engine.util.Vector2D;


/**
 * An abstract base class representing a Convex Polygon and suitable for use with
 * the Separating Axis Theorem method of Collision Detection.
 * 
 * 
 * @author Scott Wallace
 *
 */
public abstract class ConvexPolygon {

	protected static final double NO_OVERLAP = Double.MIN_VALUE;

	/**
	 * The offset from the 'position' to the center of rotation. For a rectangle
	 * this will be half the height and width.
	 */
	protected Vector2D offsetToRotation;

	/**
	 * The 'position' of the shape. This is really just an arbitrary origin
	 * which indicates how vertex coordinates are translated from screen
	 * coordinates. A vertex at (0,0), should be rendered at (position.getX(),
	 * position.getY())
	 */
	protected Vector2D position;

	/**
	 * Coordinates of polygon before rotation.  These are in the polygon's default
	 * coordinate system which has the origin at the upper left corner.
	 * Vertices are listed in clockwise fashion.
	 */
	protected Vector2D[] unrotatedVertexCoordinates;

	/**
	 * The number of corners or edges
	 */
	protected int nCorners;

	/**
	 * The angle of rotation (in radians).
	 */
	protected double rotation;

	/**
	 * The color in which the shape should be drawn.
	 */
	public Color color;

	protected double radius;
	/**
	 * A protected constructor, with minimal definition that does not
	 * necessarily create a fully initialized polygon.  Additional effort
	 * is needed to specify the vertices and ensure convexity.
	 * 
	 * @param nCorners
	 */
	protected ConvexPolygon(int nCorners) {
		this.nCorners = nCorners;
		unrotatedVertexCoordinates = new Vector2D[nCorners];
		color = Color.black;
	}

	/**
	 * Creates a rectangle with edges aligned with the x and y axes.
	 * 
	 * @param origin the upper left corner of the rectangle
	 * @param width the width 
	 * @param height the height
	 */
	public ConvexPolygon(Vector2D origin, double width, double height) {
		this(4);
		radius = Math.sqrt(width*width + height*height)/2.0;

		unrotatedVertexCoordinates[0] = new Vector2D(radius-width/2.0, radius-height/2.0);
		unrotatedVertexCoordinates[1] = new Vector2D(radius+width/2.0, radius-height/2.0);
		unrotatedVertexCoordinates[2] = new Vector2D(radius+width/2.0, radius+height/2.0);
		unrotatedVertexCoordinates[3] = new Vector2D(radius-width/2.0, radius+height/2.0);
		
		position = origin;
		rotation = 0;
		offsetToRotation = new Vector2D(radius, radius);
	}

	/**
	 * Creates a convex n-gon.
	 * <pre>
	 * 
	 *            Pxxx
	 *            x   xxxx
	 *            x   o    xxxV
	 *            x   xxxx 
	 *            xxxx
	 *            
	 * o - center of rotation
	 * P - position coordinate 
	 * V - a vertex in the polygon
	 * r - the radius (distance between o and V)
	 *          
	 * </pre>
	 * @param position the location of the axis-aligned bounding box's upper left corner.
	 * @param radius the radius of the circle that circumscribes the n-gon.
	 * @param n the desired number of sides/corners.
	 */
	public ConvexPolygon(Vector2D position, double radius, int n) {
		this(n);
		
		this.position = position;
		rotation = 0;
		this.radius = radius;
		offsetToRotation = new Vector2D(radius, radius);

		Vector2D rad = new Vector2D(radius, 0);
		Matrix22 rot;
		Vector2D rotrad;
		double deltaRadians = 2 * Math.PI / n;
		for (int i = 0; i < n; i++) {
			rot = new Matrix22(i * deltaRadians);
			rotrad = rot.multiply(rad);
			unrotatedVertexCoordinates[i] = rotrad.translate(offsetToRotation);
		}
		
	}
	
	public double getRotation() {
		return rotation;
	}

	public void setRotation(double radians) {
		rotation = radians;
	}

	public void setPosition(Vector2D p) {
		position = p;
	}

	public Vector2D getPosition() {
		return position;
	}


	/**
	 * 
	 * Task 1:
	 * 
	 * Returns the 'true' positions of each vertex taking into account:
	 * <ul>
	 * <li> the 'position' which is a constant translation applied to each
	 * vertex</li>
	 * <li> the 'offsetToRotation' which is the location (relative to position)
	 * about which the shape should be rotated</li>
	 * <li>the 'rotation' which is the angle at which the shape should be drawn</li>
	 * </ul>
	 * 
	 * @return the new positions of each vertex
	 */
	protected abstract Vector2D[] calculateVertexPositions();

	
	/**
	 * Task 2:
	 * 
	 * Determines whether a given point lies within the polygon. This method is
	 * used by the test environment to render your polygon from an existing texture.
	 * 
	 * 
	 * @param x the x-coordinate of the specified test point
	 * @param y the y-coordinate of the specified test point
	 * @return <code>true</code> iff the specified point lies on
	 * the inside of the polygon
	 *
	 */
	 public abstract boolean contains(double x, double y);
	
	/**
	 * Task 3:
	 * 
	 * Calculates the lefthand (outward pointing) edge normals. These should
	 * (obviously) be normal to each edge, outward facing, and in addition they
	 * should be of unit length;
	 * 
	 * Note: Edge i begins at vertex (i-1) and ends at the ith vertex Edge 0
	 * begins at vertex (nCorners - 1) and ends and vertex 0
	 * 
	 * Once this is completed, you should be able to render the markup for the
	 * convex polygons in the SATTestEnvironment
	 * 
	 * @return an array of unit length outward pointing edge normals
	 */
	abstract public Vector2D[] computeLeftHandEdgeNormals();

	/**
	 * Task 4:
	 * 
	 * Returns all possible separating axes. Clever implementations will cull
	 * unnecessary or redundant axes to save time.
	 * 
	 * @param other a second polygon whose axes should also be considered.
	 *  If this argument is <code>null</code> only the separating axes for
	 *  this polygon are returned.
	 *  
	 * @return a list of all possible separating axes.
	 */
	public abstract List<Vector2D> getPotentialSeparatingAxes(
			ConvexPolygon other);

	/**
	 * Task 5:
	 * 
	 * Checks for overlap between this ConvexPolygon and another along the
	 * specified separating axis. When the verbose flag is true various
	 * debugging information can be printed, logged, etc. When false
	 * implementations should be sure not to perform any unnecessary output.
	 * 
	 * @param other
	 *            the polygon which may be intersecting this one.
	 * @param axis
	 *            one of the potentially separating axes
	 * @return the amount of intersection along this axis or NO_OVERLAP
	 * 
	 * @see #getPotentialSeparatingAxes(ConvexPolygon)
	 */
	public abstract double intersectionTest(ConvexPolygon other,
			Vector2D axis, boolean verbose);

	/**
	 * Task 6:
	 * 
	 * Returns the direction that a shape should be moved to most quickly
	 * resolve a collision. This is along one of the separating axes (the axis
	 * in which there is minimum penetration). Note that this is not necessarily
	 * the direction that the object came from.
	 * 
	 * If no collision is detected, this method returns <code>null</code>
	 * 
	 * @param other
	 *            the other shape involved in the (potential) collision
	 * @param verbose
	 *            <code>true</code> if it is acceptable to print information
	 *            to stdout or stderr.
	 * @return the direction to move this shape to resolve the collision
	 * 
	 */
	public abstract Vector2D minPenetration(ConvexPolygon other,
			boolean verbose);

	
	/**
	 * Renders additional 'markup' that could be useful for debugging.
	 * The markup can be turned on or off in the test environment.
	 * By default, this method renders the normal vectors from the mid point of each edge.
	 * 
	 * @param rc
	 *            the game frame's rendering context
	 * @param se
	 *            a shape engine
	 */
	public void renderMarkup(RenderingContext rc, ShapeEngine se) {

		Vector2D[] midpoints = computeEdgeMidPoints();
		Vector2D[] normals = computeLeftHandEdgeNormals();
		
		for (int i = 0; i < nCorners; i++) {
			se.renderVector(rc, color, midpoints[i], normals[i].scale(25));
		}
	}

	/**
	 * Computes the mid points of each edge from its true coordinates.
	 * 
	 * @return a vector of midpoints edge i ends extends between vertex (i-1)
	 *         and vertex i
	 */
	public Vector2D[] computeEdgeMidPoints() {
		Vector2D[] vPositions = calculateVertexPositions();

		Vector2D[] midpoints = new Vector2D[nCorners];

		Vector2D startVertex = vPositions[nCorners - 1];
		Vector2D endVertex;

		for (int i = 0; i < nCorners; i++) {
			endVertex = vPositions[i];
			midpoints[i] = new Vector2D(startVertex.getX()
					+ (endVertex.getX() - startVertex.getX()) / 2.0,
					startVertex.getY()
							+ (endVertex.getY() - startVertex.getY()) / 2.0);
			startVertex = endVertex;
		}
		return midpoints;
	}

	@Override
	public String toString() {
		Vector2D[] v = calculateVertexPositions(); 
		StringBuffer sb = new StringBuffer(40);
		sb.append("ConvexPolygon (");
		sb.append(nCorners);
		sb.append(") @ (");
		sb.append(position.getX());
		sb.append(',');
		sb.append(position.getY());
		sb.append(") [");
		for (int i = 0; i < nCorners; i++) {
			sb.append('(');
			sb.append(v[i].getX());
			sb.append(',');
			sb.append(v[i].getY());
			sb.append(") ");
		}
		
		sb.append(']');
		return sb.toString();
	}
}
