package jig.misc.sat;

import java.awt.Color;
import java.awt.Font;
import java.awt.geom.AffineTransform;
import java.text.DecimalFormat;
import java.util.ArrayList;

import jig.engine.FontResource;
import jig.engine.RenderingContext;
import jig.engine.ResourceFactory;
import jig.engine.ShapeEngine;
import jig.engine.util.Vector2D;

/**
 * This SATLayer contains information about each of the shapes
 * in the test environment.
 * 
 * Data is updated periodically, although not necessarily each frame.
 * 
 * @author Scott Wallace
 *
 */
public class HUDLayer implements SATLayer {
	boolean active;
	DecimalFormat fmt1d = new DecimalFormat("0.0;-0.0");
	DecimalFormat fmt3d = new DecimalFormat("0.000;-0.000");
	FontResource font;
	SATTestEnvironment env;
	ShapeEngine shapeEngine;
	
	long timeSinceRefresh;
	long updatePeriod;
	ArrayList<String> properties;
	
	
	Vector2D reboundAxis;
	Vector2D reboundOrigin = new Vector2D(250, 350);
	
	/**
	 * Creates a new (mostly initialized) layer. Immediately after
	 * the constructor is called, the layer should be tied to its 
	 * environment using <code>setEnvironment</code>
	 * 
	 * @see #setEnvironment(SATTestEnvironment)
	 */
	public HUDLayer() {
		active = false;
		ResourceFactory f = ResourceFactory.getFactory();
		font = f.getFontResource(new Font("Sans Serif", Font.PLAIN, 12), Color.black, null);		

		timeSinceRefresh = 0;
		updatePeriod = 100;
		
		properties = new ArrayList<String>(20);
		reboundAxis = null;

	}
	
	/**
	 * Ties the layer to its environment. Calling this method
	 * also makes this layer active.
	 * 
	 * @param env the environment this layer is tied to.
	 */
	public void setEnvironment(SATTestEnvironment env) {
		active = true;
		this.env = env;
		this.shapeEngine = env.shapeEngine;
		
	}
	
	/**
	 * Renders the layer using information about the shapes
	 * from the last update.
	 */
	public void render(RenderingContext rc) {
		if (!active) return;

		AffineTransform at = AffineTransform.getTranslateInstance(50, 300);
		for (String s : properties) {
			at.translate(0, 16);
			font.render(s, rc, at);
		}

		
		at = AffineTransform.getTranslateInstance(250, 500);
		font.render("Polygon Factory: " + env.getPolygonFactoryName(), rc, at);
		if (reboundAxis != null)
			shapeEngine.renderVector(rc, Color.blue, 
				reboundOrigin, reboundAxis.scale(25) );

	}
	
	/**
	 * @return <code>true</code>. This is an overlay layer.
	 */
	public boolean isOverlay() { return true; }

	/**
	 * Refreshes HUD Data
	 * 
	 * @param verbose a flag to indicate if debugging data should
	 * be sent.  When called from 'update' the verbose flag is <code>false</code>
	 * when called due to a key press it is <code>true</code>.
	 * 
	 */
	public void refresh(boolean verbose) {
		
		// reset the timeSinceRefresh
		timeSinceRefresh = 0;
		
		Vector2D v;
		
		properties.clear();
		v = env.vp1.getPosition();
		properties.add("Shape 1 Properties:");
		properties.add("Position: " + fmt1d.format(v.getX()) 
								+ ',' + fmt1d.format(v.getY()));
		properties.add("Rotation: " + fmt3d.format(env.vp1.getRotation()));
		
		
		v = env.vp2.getPosition();
		properties.add("");
		properties.add("Shape 2 Properties:");
		properties.add("Position: " + fmt1d.format(v.getX()) 
				+ ',' + fmt1d.format(v.getY()));
		properties.add("Rotation: " + fmt3d.format(env.vp2.getRotation()));	

		reboundAxis = env.vp1.minPenetration(env.vp2);
		properties.add("");
		if (reboundAxis == null) 
			properties.add("Collison: NO");
		else
			properties.add("Collison: YES");
		

	}
	
	
	/**
	 * After a predefined period of time, the HUD Data will be 
	 * refreshed. This won't necessarily happen each frame.
	 */
	public void update(long deltaMs) {
		timeSinceRefresh += deltaMs;
		
		// is it time to update?
		if (timeSinceRefresh > updatePeriod) {
			refresh(false);	
		}
	}

	/**
	 * @return true if the HUD is active
	 */
	public boolean isActive() {
		return active;
	}

	/**
	 * Sets the activation state for this layer.
	 * The layer should not be made active until after <code>setEnvironment</code>
	 * has been called to finish initialization.
	 * 
	 * @param a the new activation state
	 * @see #setEnvironment(SATTestEnvironment)
	 */
	public void setActivation(boolean a) {
		active = a;
	}
	
}