package jig.misc.sat;

import java.awt.Color;
import java.awt.event.KeyEvent;
import java.util.Iterator;
import java.util.LinkedList;

import javax.imageio.spi.ServiceRegistry;

import jig.engine.RenderingContext;
import jig.engine.ResourceFactory;
import jig.engine.ShapeEngine;
import jig.engine.ViewableLayer;
import jig.engine.hli.StaticScreenGame;
import jig.engine.j2d.J2DShapeEngine;
import jig.engine.physics.vpe.VanillaPolygon;
import jig.engine.util.Vector2D;

/**
 * A Collision Detection Test Skeleton.
 * 
 * The SATTestEnvironment looks for two types of services that can be provided
 * either by the skeleton code base (the jar file provided for students) or by
 * the student's submissions. Services are advertised in the META-INF/services
 * directory through the standard service provider protocol.
 * 
 * The SATTestEnvironment looks for and automatically loads the following
 * services:
 * 
 * <ul>
 * <li> SATLayer instances (visualization services that add information to the
 * test environment). These layers are rendered on top of the default display, so
 * be sure not to obscure existing information.</li>
 * 
 * <li>PolygonFactory instances (factories that create the polygons used by the
 * test environment) </li>
 * </ul>
 * 
 * Key Controls:
 * <ul>
 * <li><code>W,A,S,D</code> - Move shape 1 up, left, down, right respectively. </li>
 * <li><code>Q,R</code> - Rotate shape 1 left, right respectively. </li>
 * <li><code>I,J,K,L</code> - Move shape 2 up, left, down, right respectively. </li>
 * <li><code>U,O</code> - Rotate shape 2 left, right respectively. </li>
 * <li><code>1,2,3</code> - Make shape 1 a 3,4,5 sided regular polygon respectively. </li>
 * <li><code>4,5,6</code> - Make shape 2 a 3,4,5 sided regular polygon respectively. </li>
 * <li><code>-,=</code> - Make shape 1 smaller or bigger respectively. </li>
 * <li><code>Z</code> - Refresh any existing SATLayers including the HUDLayer which performs the intersection test. Also dump collision information to stdout. </li>
 * <li><code>, .</code> - Switch between existing PolygonFactory objects (current factory is displayed at the bottom of the screen). </li>
 * <li><code>space</code> - Toggle whether the <code>renderMarkup</code> method is called on the polygons. </li>
 * </ul>
 * 
 * Note that the collision detection is not performed every frame. Rather it is performed 
 * approximately every 100ms (a parameter set in the HUDLayer),
 * this means there may be a small delay between movement and visual response.
 * 
 * 
 * 
 * @author Scott Wallace
 * 
 */
public class SATTestEnvironment extends StaticScreenGame {

	public static final String SPRITE_SHEET = "resources/textures.png";
	public static final String SPRITE_SHEET_XML = "resources/textures.xml";
	public static final int HEIGHT = 600;
	public static final int WIDTH = 800;
	public static final Vector2D UP = new Vector2D(0, -1);
	public static final Vector2D DOWN = new Vector2D(0, 1);
	public static final Vector2D LEFT = new Vector2D(-1, 0);
	public static final Vector2D RIGHT = new Vector2D(1, 0);

	private boolean vkisDown[];
	private static final int Z = 0;
	private static final int PERIOD = 1;
	private static final int SPACE = 2;
	private static final int N_TYPED_KEYS = 3;

	private int STD_SHAPE_RADIUS = 32;
	private int MAX_SHAPE_RADIUS = 32;
	private int MIN_SHAPE_RADIUS = 16;

	public J2DShapeEngine shapeEngine;
	public VanillaPolygon vp1;
	public VanillaPolygon vp2;

	private boolean renderMarkup;
	private final Vector2D shape1InitLoc = new Vector2D(100, 100);
	private final Vector2D shape2InitLoc = new Vector2D(200, 100);

	private Vector2D ulc = new Vector2D(0, 0);
	private Vector2D quadrant = new Vector2D(400, 300);
	private Color straw = new Color(192, 192, 128);

	private PolygonFactory factory;
	private LinkedList<PolygonFactory> polygonProviders;

	public SATTestEnvironment() {
		super(WIDTH, HEIGHT, false);

		//load the textures
		ResourceFactory.getFactory().loadSheet(SPRITE_SHEET, SPRITE_SHEET_XML);
		
		shapeEngine = new J2DShapeEngine(gameframe);
		TestVanillaPolygon.shapeEngine = shapeEngine;
		// initialize array for identifying key 'typed' events
		vkisDown = new boolean[N_TYPED_KEYS];
		for (int i = 0; i < N_TYPED_KEYS; i++)
			vkisDown[i] = false;

		// look up service providers for: PolygonFactory and SATLayer
		polygonProviders = new LinkedList<PolygonFactory>();
		for (Iterator<PolygonFactory> f = ServiceRegistry
				.lookupProviders(PolygonFactory.class); f.hasNext();) {
			polygonProviders.add(f.next());
		}
		if (polygonProviders.size() == 0) {
			throw new IllegalStateException(
					"Couldn't find a valid PolygonFactory service...is there a META-INFO/services entry?");
		}
		nextPolygonFactory();

		// SATLayer services
		SATLayer layer;
		for (Iterator<SATLayer> il = ServiceRegistry
				.lookupProviders(SATLayer.class); il.hasNext();) {
			layer = il.next();
			layer.setEnvironment(this);
			gameObjectLayers.add(layer);
			System.out.println(" Loaded new SATLayer: "
					+ layer.getClass().getName());
		}
	}

	/**
	 * Sets the current polygon factory to the next factory and then recreates
	 * the shapes.
	 */
	final void nextPolygonFactory() {

		// set factory to head of list, and then move that element to the tail.
		factory = polygonProviders.remove();
		polygonProviders.add(factory);

		System.out.println("Factory is:" + factory.getClass().toString());
		// recreate the shapes with the new polygon factory.

		vp1 = new TestVanillaPolygon(SPRITE_SHEET + "#textures");		
		vp1.setPosition(shape1InitLoc);
		vp1.setBackingShape(factory.createNGon(new Vector2D(0,0), 32, 3));
		vp1.setFrame(3);
		

		vp2 = new TestVanillaPolygon(SPRITE_SHEET + "#textures");	
		vp2.setPosition(shape2InitLoc);
		vp2.setBackingShape(factory.createRectangle(new Vector2D(0,0), 60, 28));
		vp2.setFrame(0);
	}

	/**
	 * Gets user input to control the movement and display of the shapes.
	 * 
	 */
	@Override
	public void update(long deltaMs) {
		super.update(deltaMs);

		int sides = 0;
		if (keyboard.isPressed(KeyEvent.VK_1)) sides = 3;
		if (keyboard.isPressed(KeyEvent.VK_2)) sides = 4;
		if (keyboard.isPressed(KeyEvent.VK_3)) sides = 5;
		if (sides > 0) {
			vp1.setBackingShape(factory.createNGon(shape1InitLoc, 
									STD_SHAPE_RADIUS, sides));
		}
		if (keyboard.isPressed(KeyEvent.VK_4)) sides = 3;
		if (keyboard.isPressed(KeyEvent.VK_5)) sides = 4;
		if (keyboard.isPressed(KeyEvent.VK_6)) sides = 5;
		if (sides > 0) {
			vp2.setBackingShape(factory.createNGon(shape2InitLoc, 
									STD_SHAPE_RADIUS, sides));
		}

		
		// dpx: delta Pixels -- essentially the velocity of the shapes
		double dpx = deltaMs / 20.0;

		// drd: delta radians -- essentially the angular velocity of the shapes
		double drd = deltaMs / 200.0;

		/**
		 * Controls for Shape 1
		 */
		Vector2D p = vp1.getPosition();
		double vp1r = vp1.getRadius();
		if (keyboard.isPressed(KeyEvent.VK_W)) {
			if (p.getY() > vp1r )
				vp1.setPosition(p.translate(UP.scale(dpx)));
		}
		if (keyboard.isPressed(KeyEvent.VK_A)) {
			if (p.getX() > vp1r )
				vp1.setPosition(p.translate(LEFT.scale(dpx)));
		}
		if (keyboard.isPressed(KeyEvent.VK_D)) {
			if (p.getX() < quadrant.getX() - vp1r)
				vp1.setPosition(p.translate(RIGHT.scale(dpx)));
		}
		if (keyboard.isPressed(KeyEvent.VK_S)) {
			if (p.getY() < quadrant.getY() - vp1r)
				vp1.setPosition(p.translate(DOWN.scale(dpx)));
		}
		if (keyboard.isPressed(KeyEvent.VK_Q)) {
			vp1.setRotation(vp1.getRotation() - drd);
		}
		if (keyboard.isPressed(KeyEvent.VK_E)) {
			vp1.setRotation(vp1.getRotation() + drd);
		}
		if (keyboard.isPressed(KeyEvent.VK_MINUS) && 
				vp1r > MIN_SHAPE_RADIUS) {
			vp1.setBackingShape(factory.createNGon(vp1.getPosition(), 
									vp1r-1, vp1.getNSides()));
		}
		if (keyboard.isPressed(KeyEvent.VK_EQUALS) && 
				vp1r < MAX_SHAPE_RADIUS) {
			vp1.setBackingShape(factory.createNGon(vp1.getPosition(), 
									vp1r+1, vp1.getNSides()));
		}
		/**
		 * Controls for Shape 2
		 */
		p = vp2.getPosition();
		double vp2r = vp2.getRadius();
		if (keyboard.isPressed(KeyEvent.VK_I)) {
			if (p.getY() > vp2r)
				vp2.setPosition(p.translate(UP.scale(dpx)));
		}
		if (keyboard.isPressed(KeyEvent.VK_J)) {
			if (p.getX() > vp2r)
				vp2.setPosition(p.translate(LEFT.scale(dpx)));
		}
		if (keyboard.isPressed(KeyEvent.VK_L)) {
			if (p.getX() < quadrant.getX() - vp2r)
				vp2.setPosition(p.translate(RIGHT.scale(dpx)));
		}
		if (keyboard.isPressed(KeyEvent.VK_K)) {
			if (p.getY() < quadrant.getY() - vp2r)
				vp2.setPosition(p.translate(DOWN.scale(dpx)));
		}
		if (keyboard.isPressed(KeyEvent.VK_U)) {
			vp2.setRotation(vp2.getRotation() - drd);
		}
		if (keyboard.isPressed(KeyEvent.VK_O)) {
			vp2.setRotation(vp2.getRotation() + drd);
		}


		// look for 'key typed' events
		if (keyboard.isPressed(KeyEvent.VK_Z)) {
			vkisDown[Z] = true;
		} else {
			if (vkisDown[Z]) {
				// it was down, now it's up!
				for (ViewableLayer l : gameObjectLayers) {
					if (l instanceof SATLayer) {
						((SATLayer) l).refresh(true);
					}
				}
				System.out.println("Shape 1:" + vp1.getShape().toString());
				System.out.println("Shape 2:" + vp2.getShape().toString());
				Vector2D c = vp1.minPenetration(vp2, true);
				System.out.println("Min Penetration: " + (c==null?"-- none --":c.toString()));
				vkisDown[Z] = false;
			}

		}

		if (keyboard.isPressed(KeyEvent.VK_PERIOD)) {
			vkisDown[PERIOD] = true;
		} else {
			if (vkisDown[PERIOD]) {
				// it was down, now it's up!
				nextPolygonFactory();
				vkisDown[PERIOD] = false;
			}

		}
		if (keyboard.isPressed(KeyEvent.VK_SPACE)) {
			vkisDown[SPACE] = true;
		} else {
			if (vkisDown[SPACE]) {
				// it was down, now it's up!
				renderMarkup = !renderMarkup;
				vp1.renderMarkup(renderMarkup);
				vp2.renderMarkup(renderMarkup);
				vkisDown[SPACE] = false;
			}

		}

	}

	/**
	 * Render the shapes, and possible shape markup
	 */
	@Override
	public void render(RenderingContext rc) {

		shapeEngine.renderFilledRectangle(rc, straw, ulc, quadrant);
		vp2.render(rc);
		vp1.render(rc);

		super.render(rc);

	}

	public String getPolygonFactoryName() {
		return factory.getClass().getName();
	}

	public PolygonFactory getFactory() {
		return factory;
	}

	public static final void main(String[] args) {
		//J2DResourceFactory.makeCurrentResourceFactory();

		SATTestEnvironment testEnv = new SATTestEnvironment();
		testEnv.run();
	}
}

class TestVanillaPolygon extends VanillaPolygon {

	static ShapeEngine shapeEngine;
	public TestVanillaPolygon(String rsc) { super(rsc); }
	
	@Override
	public void update(long deltaMs) {
		// TODO Auto-generated method stub
		
	}
	@Override
	public void render(RenderingContext rc) {
		super.render(rc);
		
		if (renderMarkup)
			shape.renderMarkup(rc, shapeEngine);
	}
	
}